package net.nosial.socialclient.classes;


import net.nosial.socialclient.exceptions.CryptographyException;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import java.nio.charset.StandardCharsets;
import java.security.*;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.X509EncodedKeySpec;
import java.util.ArrayList;
import java.util.Base64;
import java.util.List;

public final class Cryptography
{
    private static final int TIME_BLOCK = 60;
    private static final int KEY_SIZE = 2048;
    private static final String ALGORITHM = "RSA";
    private static final String SIGNATURE_ALGORITHM = "SHA256withRSA";
    private static final String CIPHER_TRANSFORMATION = "RSA/ECB/OAEPWithSHA-1AndMGF1Padding";
    private static final int MAX_ENCRYPT_BLOCK_SIZE = 214;
    private static final int MAX_DECRYPT_BLOCK_SIZE = 256;

    /**
     * Returns the KeyFactory object for the specified algorithm.
     *
     * @return the KeyFactory object for the specified algorithm
     * @throws NoSuchAlgorithmException if the requested algorithm is not available
     */
    private static KeyFactory getKeyFactory() throws NoSuchAlgorithmException
    {
        return KeyFactory.getInstance(ALGORITHM);
    }

    /**
     * Generates a key pair using the RSA algorithm.
     *
     * @return the generated key pair
     * @throws CryptographyException if there is an error generating the key pair
     */
    public static KeyPair generateKeyPair() throws CryptographyException
    {
        try
        {
            KeyPairGenerator keyPairGenerator = KeyPairGenerator.getInstance(ALGORITHM);
            keyPairGenerator.initialize(KEY_SIZE);
            return keyPairGenerator.generateKeyPair();
        }
        catch (NoSuchAlgorithmException e)
        {
            throw new CryptographyException("Failed to generate key pair, RSA algorithm not found", e);
        }
    }

    /**
     * Signs the given content using the provided private key.
     *
     * @param content     the content to be signed
     * @param privateKey  the private key used for signing
     * @return the base64 encoded signature of the content
     * @throws CryptographyException if an error occurs in the cryptography process
     */
    public static String signContent(String content, String privateKey) throws CryptographyException
    {
        return signContent(content, importPrivateKey(privateKey));
    }

    /**
     * Signs the given content using the provided private key.
     *
     * @param content     the content to be signed
     * @param privateKey  the private key used for signing
     * @return the base64 encoded signature of the content
     * @throws CryptographyException if an error occurs in the cryptography process
     */
    public static String signContent(String content, PrivateKey privateKey) throws CryptographyException
    {
        try
        {
            Signature signature = Signature.getInstance(SIGNATURE_ALGORITHM);
            signature.initSign(privateKey);
            signature.update(sha1(content).getBytes(StandardCharsets.UTF_8));
            byte[] signatureBytes = signature.sign();

            // Base64 encode the signature
            return Base64.getEncoder().encodeToString(signatureBytes);
        }
        catch (NoSuchAlgorithmException | InvalidKeyException | SignatureException e)
        {
            throw new CryptographyException("Failed to sign content", e);
        }
    }

    private static String bytesToHex(byte[] hash)
    {
        StringBuilder hexString = new StringBuilder(2 * hash.length);
        for (byte b : hash)
        {
            String hex = Integer.toHexString(0xff & b);
            if (hex.length() == 1)
            {
                hexString.append('0');
            }
            hexString.append(hex);
        }

        return hexString.toString();
    }

    /**
     * Verifies the content using the provided signature and public key.
     *
     * @param content   the content to be verified
     * @param signature the signature of the content
     * @param publicKey the public key used for verification
     * @return true if the content is verified, false otherwise
     * @throws CryptographyException if an error occurs in the cryptography process
     */
    public static boolean verifyContent(String content, String signature, String publicKey) throws CryptographyException
    {
        return verifyContent(content, signature, importPublicKey(publicKey));
    }

    /**
     * Verifies the signature of the given content using the provided public key.
     *
     * @param content    the content to be verified
     * @param signature  the signature to be verified
     * @param publicKey  the public key used for verification
     * @return true if the signature is verified, false otherwise
     * @throws CryptographyException if there is an error during the verification process
     */
    public static boolean verifyContent(String content, String signature, PublicKey publicKey) throws CryptographyException
    {
        try
        {
            final Signature sign = Signature.getInstance(SIGNATURE_ALGORITHM);
            sign.initVerify(publicKey);
            sign.update(sha1(content).getBytes(StandardCharsets.UTF_8));

            return sign.verify(Base64.getDecoder().decode(signature));
        }
        catch (NoSuchAlgorithmException | InvalidKeyException | SignatureException e)
        {
            throw new CryptographyException("Failed to verify content", e);
        }
    }

    /**
     * Generates a temporary signature for the given content using the provided private key.
     *
     * @param content     the content to be signed
     * @param privateKey  the private key used for signing
     * @return the base64 encoded signature of the content
     * @throws CryptographyException if an error occurs in the cryptography process
     */
    public static String temporarySignature(String content, String privateKey) throws CryptographyException
    {
        return signContent(String.format("%s|%d", content, (System.currentTimeMillis() / 1000 / TIME_BLOCK)), privateKey);
    }

    /**
     * Generates a temporary signature for the given content using the provided private key.
     *
     * @param content     the content to be signed
     * @param privateKey  the private key used for signing
     * @return the base64 encoded signature of the content
     * @throws CryptographyException if an error occurs in the cryptography process
     */
    public static String temporarySignature(String content, PrivateKey privateKey) throws CryptographyException
    {
        return signContent(String.format("%s|%d", content, (System.currentTimeMillis() / 1000 / TIME_BLOCK)), privateKey);
    }

    /**
     * Verifies the temporary signature of the given content using the specified public key and number of frames.
     *
     * @param content    The content to verify the signature for. Must not be null.
     * @param signature  The signature to verify. Must not be null.
     * @param publicKey  The public key used for verification. Must not be null.
     * @param frames     The number of frames to use for verification. Must be a positive integer.
     * @return true if the signature is valid, false otherwise.
     * @throws CryptographyException if an error occurs during verification.
     */
    public static boolean verifyTemporarySignature(String content, String signature, String publicKey, int frames) throws CryptographyException
    {
        return verifyTemporarySignature(content, signature, importPublicKey(publicKey), frames);
    }

    /**
     * Verifies the temporary signature of the given content using the specified public key and number of frames.
     *
     * @param content    The content to verify the signature for. Must not be null.
     * @param signature  The signature to verify. Must not be null.
     * @param publicKey  The public key used for verification. Must not be null.
     * @param frames     The number of frames to use for verification. Must be a positive integer.
     * @return true if the signature is valid, false otherwise.
     * @throws CryptographyException if an error occurs during verification.
     */
    public static boolean verifyTemporarySignature(String content, String signature, PublicKey publicKey, int frames) throws CryptographyException
    {
        if (frames <= 0) frames = 1;
        long currentTime = System.currentTimeMillis() / 1000 / TIME_BLOCK;

        for (int i = 0; i < frames; i++)
        {
            if (verifyContent(String.format("%s|%d", content, currentTime - i), signature, publicKey))
            {
                return true;
            }
        }
        return false;
    }

    /**
     * Encrypts the given content using the provided public key.
     *
     * @param content    the content to be encrypted
     * @param publicKey  the public key used for encryption
     * @return the encrypted content as a String
     * @throws CryptographyException if an error occurs during the encryption process
     */
    public static String encrypt(String content, String publicKey) throws CryptographyException
    {
        return encrypt(content, importPublicKey(publicKey));
    }

    /**
     * Encrypts the given content using the provided public key.
     *
     * @param content    the content to be encrypted
     * @param publicKey  the public key used for encryption
     * @return the encrypted content as a Base64 encoded string
     * @throws CryptographyException if an error occurs during the encryption process
     */
    public static String encrypt(String content, PublicKey publicKey) throws CryptographyException
    {
        try
        {
            Cipher cipher = Cipher.getInstance(CIPHER_TRANSFORMATION);
            cipher.init(Cipher.ENCRYPT_MODE, publicKey);

            byte[] data = content.getBytes();
            int inputLength = data.length;
            List<Byte> outputList = new ArrayList<>();

            for (int i = 0; i < inputLength; i += MAX_ENCRYPT_BLOCK_SIZE)
            {
                int length = Math.min(inputLength - i, MAX_ENCRYPT_BLOCK_SIZE);
                byte[] encryptedBlock = cipher.doFinal(data, i, length);
                for (byte b : encryptedBlock)
                {
                    outputList.add(b);
                }
            }

            byte[] output = new byte[outputList.size()];
            for (int i = 0; i < outputList.size(); i++)
            {
                output[i] = outputList.get(i);
            }

            return Base64.getEncoder().encodeToString(output);
        }
        catch (NoSuchAlgorithmException | NoSuchPaddingException | InvalidKeyException | IllegalBlockSizeException | BadPaddingException e)
        {
            throw new CryptographyException("Failed to encrypt content", e);
        }
    }

    /**
     * Decrypts the given content using the provided private key.
     *
     * @param content    the encrypted content to be decrypted
     * @param privateKey the private key used for decryption
     * @return the decrypted content as a String
     * @throws CryptographyException if an error occurs during the decryption process
     */
    public static String decrypt(String content, String privateKey) throws CryptographyException
    {
        return decrypt(content, importPrivateKey(privateKey));
    }

    /**
     * Decrypts the given content using the provided private key.
     *
     * @param content the encrypted content to be decrypted
     * @param privateKey the private key used for decryption
     * @return the decrypted content as a String
     * @throws CryptographyException if an error occurs during the decryption process
     */
    public static String decrypt(String content, PrivateKey privateKey) throws CryptographyException
    {
        try
        {
            Cipher cipher = Cipher.getInstance(CIPHER_TRANSFORMATION);
            cipher.init(Cipher.DECRYPT_MODE, privateKey);

            byte[] data = Base64.getDecoder().decode(content);
            int inputLength = data.length;
            List<Byte> outputList = new ArrayList<>();

            for (int i = 0; i < inputLength; i += MAX_DECRYPT_BLOCK_SIZE)
            {
                int length = Math.min(inputLength - i, MAX_DECRYPT_BLOCK_SIZE);
                byte[] decryptedBlock = cipher.doFinal(data, i, length);
                for (byte b : decryptedBlock)
                {
                    outputList.add(b);
                }
            }

            byte[] output = new byte[outputList.size()];
            for (int i = 0; i < outputList.size(); i++)
            {
                output[i] = outputList.get(i);
            }

            return new String(output);
        }
        catch (NoSuchAlgorithmException | NoSuchPaddingException | InvalidKeyException | IllegalBlockSizeException | BadPaddingException e)
        {
            throw new CryptographyException("Failed to decrypt content", e);
        }
    }

    /**
     * Exports the given private key as a Base64 encoded string.
     *
     * @param privateKey the private key to be exported
     * @return the Base64 encoded string representation of the private key
     */
    public static String exportPrivateKey(PrivateKey privateKey)
    {
        return Base64.getEncoder().encodeToString(privateKey.getEncoded());
    }

    /**
     * Exports the given private key as a Base64 encoded string.
     *
     * @param keyPair the private key to be exported
     * @return the Base64 encoded string representation of the private key
     */
    public static String exportPrivateKey(KeyPair keyPair)
    {
        return exportPrivateKey(keyPair.getPrivate());
    }

    /**
     * Exports the given public key as a Base64 encoded string.
     *
     * @param publicKey the public key to be exported
     * @return a Base64 encoded string representation of the public key
     */
    public static String exportPublicKey(PublicKey publicKey)
    {
        return Base64.getEncoder().encodeToString(publicKey.getEncoded());
    }

    /**
     * Exports the given public key as a Base64 encoded string.
     *
     * @param keyPair the public key to be exported
     * @return a Base64 encoded string representation of the public key
     */
    public static String exportPublicKey(KeyPair keyPair)
    {
        return exportPublicKey(keyPair.getPublic());
    }

    /**
     * Imports a public key from a Base64 encoded string representation.
     *
     * @param publicKey the Base64 encoded string representation of the public key
     * @return the imported PublicKey object
     * @throws CryptographyException if there is an error importing the public key
     */
    public static PublicKey importPublicKey(String publicKey) throws CryptographyException
    {
        try
        {
            byte[] keyBytes = Base64.getDecoder().decode(publicKey);
            X509EncodedKeySpec spec = new X509EncodedKeySpec(keyBytes);
            return getKeyFactory().generatePublic(spec);
        }
        catch (InvalidKeySpecException | NoSuchAlgorithmException | IllegalArgumentException e)
        {
            throw new CryptographyException("Failed to import public key", e);
        }
    }

    /**
     * Imports a private key from a Base64 encoded string representation.
     *
     * @param privateKey the Base64 encoded string representation of the private key
     * @return the imported PrivateKey object
     * @throws CryptographyException if there is an error importing the private key
     */
    public static PrivateKey importPrivateKey(String privateKey) throws CryptographyException
    {
        try
        {
            byte[] keyBytes = Base64.getDecoder().decode(privateKey);
            PKCS8EncodedKeySpec spec = new PKCS8EncodedKeySpec(keyBytes);
            return getKeyFactory().generatePrivate(spec);
        }
        catch (InvalidKeySpecException | NoSuchAlgorithmException | IllegalArgumentException e)
        {
            throw new CryptographyException("Failed to import private key", e);
        }
    }

    public static String sha1(String data) throws NoSuchAlgorithmException
    {
        MessageDigest digest = MessageDigest.getInstance("SHA-1");
        byte[] hash = digest.digest(data.getBytes(StandardCharsets.UTF_8));
        return bytesToHex(hash);
    }

}
