    package net.nosial.socialclient.classes;
    
    import com.fasterxml.jackson.core.JsonProcessingException;
    import com.fasterxml.jackson.core.type.TypeReference;
    import com.fasterxml.jackson.databind.JsonNode;
    import com.fasterxml.jackson.databind.ObjectMapper;
    import com.squareup.okhttp.*;
    import net.nosial.socialclient.abstracts.RpcResult;
    import net.nosial.socialclient.exceptions.CryptographyException;
    import net.nosial.socialclient.exceptions.ResolutionException;
    import net.nosial.socialclient.objects.ResolvedServer;
    import net.nosial.socialclient.objects.RpcRequest;
    
    import java.io.IOException;
    import java.security.NoSuchAlgorithmException;
    import java.security.PrivateKey;
    import java.security.PublicKey;
    import java.util.ArrayList;
    import java.util.List;
    import java.util.Map;
    
    public class RpcClient
    {
        private final static ObjectMapper OBJECT_MAPPER = new ObjectMapper();
        private final static MediaType MEDIA_TYPE_JSON = MediaType.parse("application/json; charset=utf-8");
    
        private final String domain;
        private final String endpoint;
        private final PublicKey serverPublicKey;
        private final OkHttpClient httpClient;
    
        private String clientName = "SocialClient Java";
        private String clientVersion = "1.0";
    
        protected String sessionUuid;
        protected PrivateKey privateKey;
    
        /**
         * Initializes a new instance of the RpcClient class using the specified domain.
         * This constructor resolves the domain to obtain the RPC endpoint and public key.
         *
         * @param domain The domain to be resolved for the RPC endpoint and public key.
         * @throws ResolutionException If the domain cannot be resolved or if required information is missing.
         * @throws CryptographyException If an error occurs while importing the public key.
         */
        public RpcClient(String domain) throws ResolutionException, CryptographyException
        {
            this.domain = domain;
            this.httpClient = new OkHttpClient();
    
            // Resolve the domain to get the endpoint and public key
            ResolvedServer resolvedServer = Resolver.resolveDomain(domain);
    
            this.endpoint = resolvedServer.getEndpoint();
            this.serverPublicKey = Cryptography.importPublicKey(resolvedServer.getPublicKey());
            this.sessionUuid = null;
            this.privateKey = null;
        }
    
        /**
         * Constructs an RpcClient instance with the specified endpoint and public key.
         *
         * @param endpoint The endpoint to which RPC requests will be sent.
         * @param serverPublicKey The public key used for cryptographic operations with the RPC server.
         */
        public RpcClient(String endpoint, PublicKey serverPublicKey)
        {
            this.domain = null;
            this.endpoint = endpoint;
            this.serverPublicKey = serverPublicKey;
            this.httpClient = new OkHttpClient();
            this.sessionUuid = null;
            this.privateKey = null;
        }
    
        /**
         * Retrieves the domain associated with this RpcClient instance.
         *
         * @return the domain as a string.
         */
        public String getDomain()
        {
            return this.domain;
        }
    
        /**
         * Returns the RPC endpoint for the client.
         *
         * @return the endpoint as a String.
         */
        public String getEndpoint()
        {
            return this.endpoint;
        }
    
        /**
         * Returns the public key associated with the RpcClient instance.
         *
         * @return the public key as a PublicKey object.
         */
        public PublicKey getServerPublicKey()
        {
            return this.serverPublicKey;
        }
    
        /**
         * Retrieves the session UUID associated with the RpcClient instance.
         *
         * @return the session UUID as a string.
         */
        public String getSessionUuid()
        {
            return this.sessionUuid;
        }
    
        /**
         * Sets the session UUID and imports the provided private key.
         *
         * @param sessionUuid the unique identifier for this session
         * @param privateKey the private key to be imported
         * @throws CryptographyException if an error occurs while importing the private key
         */
        protected void setSession(String sessionUuid, String privateKey) throws CryptographyException
        {
            this.sessionUuid = sessionUuid;
            this.privateKey = Cryptography.importPrivateKey(privateKey);
        }
    
        /**
         * Sets the session details for the RPC client.
         *
         * @param sessionUuid the UUID of the session
         * @param privateKey the private key associated with the session
         */
        protected void setSession(String sessionUuid, PrivateKey privateKey)
        {
            this.sessionUuid = sessionUuid;
            this.privateKey = privateKey;
        }
    
        /**
         * Retrieves the client name associated with this RpcClient instance.
         *
         * @return the client name as a string.
         */
        public String getClientName()
        {
            return clientName;
        }
    
        /**
         * Sets the name of the client.
         *
         * @param clientName The name to be set for the client.
         * @throws NullPointerException if the provided*/
        public void setClientName(String clientName)
        {
            if(clientName == null)
            {
                throw new NullPointerException("Client name cannot be null");
            }
    
            this.clientName = clientName;
        }
    
        /**
         * Returns the client's version.
         *
         * @return the client version as a string.
         */
        public String getClientVersion()
        {
            return clientVersion;
        }
    
        /**
         * Sets the client version for the RpcClient instance.
         *
         * @param clientVersion The version of the client software to be set.
         * @throws NullPointerException if the provided clientVersion*/
        public void setClientVersion(String clientVersion)
        {
            if(clientVersion == null)
            {
                throw new NullPointerException("Client version cannot be null");
            }
    
            this.clientVersion = clientVersion;
        }
    
        /**
         * Clears the current session by setting the sessionUuid and privateKey fields to null.
         * This method effectively logs out the user from the current session.
         */
        public void clearSession()
        {
            this.sessionUuid = null;
            this.privateKey = null;
        }
    
        /**
         * Sends an RPC request with the provided JSON data and returns a list of RpcResult objects.
         *
         * @param jsonData the JSON-formatted string representing the RPC request data.
         * @return a list of RpcResult objects representing the response(s) from the RPC server.
         * @throws RuntimeException if there is an error during the request or response processing.
         */
        @SuppressWarnings("unchecked")
        public List<RpcResult> sendRequest(String jsonData)
        {
            final Request request = new Request.Builder()
            {{
                 this.url(endpoint);
                 this.post(RequestBody.create(MEDIA_TYPE_JSON, jsonData));
                 this.addHeader("Client-Name", clientName);
                 this.addHeader("Client-Version", clientVersion);
    
                 if(sessionUuid != null)
                 {
                     // We are seeing the session UUID correctly.
                    this.addHeader("Session-UUID", sessionUuid);
                 }
    
                 if(privateKey != null)
                 {
                     try
                     {
                         // NOTE: Signature looks okay, it's base64 encoded.
                         this.addHeader("Signature", Cryptography.signContent(jsonData, privateKey));
                     }
                     catch(CryptographyException e)
                     {
                         throw new RuntimeException(e);
                     }
                 }
            }}.build();
    
            try
            {
                final Response response = this.httpClient.newCall(request).execute();
                final String responseString = response.body().string();
    
                if (!response.isSuccessful())
                {
                    if(!responseString.isEmpty())
                    {
                        throw new RuntimeException(responseString);
                    }
    
                    throw new RuntimeException("Failed to send request: " + response.code());
                }
    
                if(response.code() == 204)
                {
                    // The response is empty
                    return new ArrayList<>();
                }
    
                if(responseString.isEmpty())
                {
                    throw new RuntimeException("The request was successful but the server did not indicate a empty response");
                }
    
                Object decoded = decode(responseString);
    
                // Singular object response
                if(decoded instanceof List<?>)
                {
                    List<Map<String, Object>> responseList = (List<Map<String, Object>>) decoded;
                    List<RpcResult> results = new ArrayList<>(responseList.size());
                    for(Map<String, Object> responseMap : responseList)
                    {
                        results.add(RpcResult.fromMap(responseMap));
                    }
    
                    return results;
                }
    
                if(decoded instanceof Map<?, ?>)
                {
                    // Singular object response
                    List<RpcResult> results = new ArrayList<>(1);
                    results.add(RpcResult.fromMap((Map<String, Object>) decoded));
                    return results;
                }
    
    
                throw new RuntimeException("Failed to decode response");
            }
            catch (IOException e)
            {
                throw new RuntimeException(e);
            }
        }
    
        /**
         * Sends an RPC request to the configured endpoint and returns the first result.
         *
         * @param request the RPC request to be sent
         * @return the first RPC result received from the response
         */
        public RpcResult sendRequest(RpcRequest request)
        {
            List<RpcResult> results = sendRequest(encode(request.toMap()));
    
            if(results.isEmpty())
            {
                return null;
            }
    
            return results.getFirst();
        }
    
        /**
         * Sends a list of RPC requests to the server and returns their results as a list of RpcResult objects.
         *
         * @param requests the list of RPC requests to be sent
         * @return a list of RpcResult objects containing the responses from the server
         */
        public List<RpcResult> sendRequests(List<RpcRequest> requests)
        {
            return sendRequest(encode(requests));
        }
    
        /**
         * Decodes a JSON string into a corresponding Java object. If the JSON string
         * represents an object, it decodes it into a Map. If the JSON string represents
         * an array, it decodes it into a List.
         *
         * @param json the JSON string to be decoded
         * @return a decoded Java object which can be a Map or a List depending on the JSON structure
         * @throws RuntimeException if the JSON string is neither an object nor an array, or if decoding fails
         */
        private static Object decode(String json)
        {
            try
            {
                JsonNode jsonNode = OBJECT_MAPPER.readTree(json);
                if (jsonNode.isObject())
                {
                    return decodeJson(OBJECT_MAPPER.readValue(json, new TypeReference<>() {}));
                }
                else if (jsonNode.isArray())
                {
                    return decodeList(OBJECT_MAPPER.readValue(json, new TypeReference<>() {}));
                }
                else
                {
                    throw new RuntimeException("JSON is neither an object nor an array");
                }
            }
            catch (JsonProcessingException e)
            {
                throw new RuntimeException("Failed to decode input", e);
            }
        }
    
        /**
         * Recursively decodes a JSON structure represented as a Map. If the value
         * of an entry is another Map, it recursively decodes that Map. If the value
         * is a List, it passes the List to the decodeList method.
         *
         * @param map the JSON structure to be decoded represented as a Map
         * @return the decoded JSON structure as a Map
         */
        @SuppressWarnings("unchecked")
        private static Map<String, Object> decodeJson(Map<String, Object> map)
        {
            for (Map.Entry<String, Object> entry : map.entrySet())
            {
                if (entry.getValue() instanceof Map)
                {
                    entry.setValue(decodeJson((Map<String, Object>) entry.getValue()));
                }
                else if (entry.getValue() instanceof List)
                {
                    entry.setValue(decodeList((List<Object>) entry.getValue()));
                }
            }
    
            return map;
        }
    
        /**
         * Recursively decodes a list of objects, transforming any nested maps or lists within it.
         *
         * @param list the list of objects to be decoded
         * @return the decoded list of objects, with all nested maps and lists transformed
         */
        @SuppressWarnings("unchecked")
        private static List<Object> decodeList(List<Object> list)
        {
            for (int i = 0; i < list.size(); i++)
            {
                final Object item = list.get(i);
                if (item instanceof Map)
                {
                    list.set(i, decodeJson((Map<String, Object>) item));
                }
                else if (item instanceof List)
                {
                    list.set(i, decodeList((List<Object>) item));
                }
            }
    
            return list;
        }
    
        /**
         * Encodes the given input object to its JSON string representation.
         *
         * @param input the object to encode
         * @return the JSON string representation of the input object
         */
        private static String encode(Object input)
        {
            try
            {
                return OBJECT_MAPPER.writeValueAsString(input);
            }
            catch(JsonProcessingException e)
            {
                throw new RuntimeException("Failed to encode input to JSON Data", e);
            }
        }
    }
