package net.nosial.socialclient.objects;

import java.util.HashMap;
import java.util.Map;

public class RpcRequest
{
    private final String method;
    private final String id;
    private final Map<String, Object> parameters;

    /**
     * Initializes a new RpcRequest object using the provided data map.
     *
     * @param data a map containing the initial values for the RpcRequest.
     *             The map should have the following expected keys:
     *             - "method": a String representing the RPC method name
     *             - "id": a String representing the unique request identifier
     *             - "params": a Map<String, Object> containing parameters for the method call
     */
    @SuppressWarnings("unchecked")
    public RpcRequest(Map<String, Object> data)
    {
        this.method = (String) data.get("method");
        this.id = (String) data.get("id");
        this.parameters = (Map<String, Object>) data.get("params");
    }

    /**
     * Constructs an RpcRequest with the specified method name, request id, and parameters.
     *
     * @param method the RPC method to be called
     * @param id the unique identifier for this request
     * @param parameters the parameters for the RPC method
     */
    public RpcRequest(String method, String id, Map<String, Object> parameters)
    {
        this.method = method;
        this.id = id;
        this.parameters = parameters;
    }

    /**
     * Constructs an RPC request with a specified method and parameters.
     * The request ID will be set to null.
     *
     * @param method the name of the method being called
     * @param parameters the parameters to be sent with the method call
     */
    public RpcRequest(String method, Map<String, Object> parameters)
    {
        this.id = null;
        this.method = method;
        this.parameters = parameters;
    }

    /**
     * Constructs an RpcRequest with the specified method and id.
     *
     * @param method The method name for the RPC request.
     * @param id The id of the RPC request.
     */
    public RpcRequest(String method, String id)
    {
        this.method = method;
        this.id = id;
        this.parameters = null;
    }

    /**
     * Constructs an RpcRequest with a specified method and no parameters or ID.
     *
     * @param method the RPC method name for the request
     */
    public RpcRequest(String method)
    {
        this.id = null;
        this.method = method;
        this.parameters = null;
    }

    /**
     * Retrieves the identifier of the RPC request.
     *
     * @return the identifier of the request.
     */
    public String getId()
    {
        return this.id;
    }

    /**
     * Gets the method name associated with the RPC request.
     *
     * @return the method name as a String.
     */
    public String getMethod()
    {
        return this.method;
    }

    /**
     * Retrieves the parameters map associated with the RPC request.
     *
     * @return a Map containing the parameters of the RPC request.
     */
    public Map<String, Object> getParameters()
    {
        return this.parameters;
    }

    /**
     * Converts the current RpcRequest object into a map representation.
     *
     * @return a map containing the method, id, and params of the RpcRequest, if present.
     */
    public Map<String, Object> toMap()
    {
        HashMap<String, Object> map = new HashMap<>();
        map.put("method", this.method);

        if(this.id != null)
        {
            map.put("id", this.id);
        }

        if(this.parameters != null)
        {
            map.put("parameters", this.parameters);
        }

        return map;
    }
}
