package net.nosial.socialclient.objects.standard;

import net.nosial.socialclient.enums.flags.PeerFlags;

import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class SelfUser
{
    private final String uuid;
    private final boolean enabled;
    private final String address;
    private final String username;
    private final String displayName;
    private final List<PeerFlags> flags;
    private final int created;

    /**
     * Constructs a SelfUser instance by populating fields from a provided data map.
     *
     * @param data a map containing user details such as uuid, username, display name, flags,
     *             enabled status, and creation time.
     */
    @SuppressWarnings("unchecked")
    public SelfUser(Map<String, Object> data)
    {
        this.uuid = (String) data.get("uuid");
        this.enabled = (boolean) data.get("enabled");
        this.username = (String) data.get("username");
        this.address = (String) data.get("address");
        this.displayName = (String) data.get("display_name");
        this.flags = ((List<String>) data.get("flags")).stream().map(PeerFlags::valueOf).collect(Collectors.toList());
        this.created = (int) data.get("created");
    }

    /**
     * Retrieves the universally unique identifier (UUID) associated with the user.
     *
     * @return the UUID as a String
     */
    public String getUuid()
    {
        return uuid;
    }

    /**
     * Checks if the user account is enabled.
     *
     * @return true if the user account is enabled; false otherwise.
     */
    public boolean isEnabled()
    {
        return enabled;
    }

    /**
     * Retrieves the username associated with this SelfUser instance.
     *
     * @return the username of the user
     */
    public String getUsername()
    {
        return username;
    }

    /**
     * Retrieves the address associated with this SelfUser instance.
     *
     * @return the address of the user
     */
    public String getAddress()
    {
        return address;
    }

    /**
     * Retrieves the display name of the user.
     *
     * @return the display name of the user
     */
    public String getDisplayName()
    {
        return displayName;
    }

    /**
     * Retrieves the list of flags associated with the user.
     *
     * @return a list of PeerFlags indicating various user permissions and statuses.
     */
    public List<PeerFlags> getFlags()
    {
        return flags;
    }

    /**
     * Returns the creation timestamp of this user.
     *
     * @return the creation timestamp as an integer
     */
    public int getCreated()
    {
        return created;
    }
}
